/*
 *  Copyright (C) 2015 Martin Blumenstingl <martin.blumenstingl@googlemail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License version 2 as published by the Free Software Foundation;
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */
#ifndef DELETEJOBTEST_H
#define DELETEJOBTEST_H

#include <QObject>
#include <kio/job.h>

class DeleteJobTest : public QObject
{
    Q_OBJECT
private Q_SLOTS:
    void initTestCase();
    void deleteFileTestCase_data() const;
    void deleteFileTestCase();
    void deleteDirectoryTestCase_data() const;
    void deleteDirectoryTestCase();

private:
    void createEmptyTestFiles(const QStringList &fileNames, const QString &path) const;
};

#endif
